/* fileVolume.C

   Copyright (C) 1999 by David Rose.

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

   To contact the original author of this program, write to:
           David Rose
           224 W. Dryden St Apt. 418
	   Glendale, CA 91202
	   USA

	   drose@lafn.org
   */

#include "system.h"

#include "fileVolume.h"
#include "mkdir_p.h"
#include "dataCell.h"
#include "database.h"
#include "queryResult.h"
#include "protect_from_shell.h"

#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <dirent.h>

FileVolume::
FileVolume(const Dataset &data) : VolumeBase(data) 
{
  setup();
}

bool FileVolume::
exists() const {
  string prove_directory = _definition + "/.";
  return (access(prove_directory.c_str(), F_OK)==0);
}

bool FileVolume::
can_read_files() const {
  string prove_directory = _definition + "/.";
  return (access(prove_directory.c_str(), R_OK | X_OK)==0);
}

bool FileVolume::
can_store_files() const {
  string prove_directory = _dirname + "/.";
  return (access(_dirname.c_str(), W_OK | X_OK)==0);
}

string FileVolume::
get_filename(const string &rel_filename) {
  return _definition + "/" + rel_filename;
}

string FileVolume::
prepare_write_file(const string &basename) {
  return _dirname + "/" + basename;
}

string FileVolume::
close_write_file(const string &full_pathname, const string &basename) {
  // This is just like store_file(), below, except we don't try to
  // move the file first--we assume it's already in the bucket.
  if (is_bucket_full()) {
    increment_bucket();
    store_file(full_pathname, basename, false);
  }

  if (_bname.empty()) {
    return basename;
  } else {
    return _bname + "/" + basename;
  }
}

string FileVolume::
store_file(const string &full_pathname, const string &basename,
	   bool leave_orig) {
  // Try to put the file in the current bucket.  If the bucket is full
  // after we do this, move to the next bucket.  Keep doing this until
  // we find a home.

  // We must try to move the file in first, and _then_ check the
  // directory size.  If we do it in the other order, we run the risk
  // of a bucket getting left overfilled because two different
  // processes moved a file into it at nearly the same time!  As it
  // is, a bucket may be momentarily overfilled, but it will never be
  // left that way.

  string old_filename = full_pathname;
  string new_filename = move_file(old_filename, basename, leave_orig);
  while (!new_filename.empty() && is_bucket_full()) {
    increment_bucket();
    old_filename = _definition + "/" + new_filename;
    new_filename = move_file(old_filename, basename, false);
  }

  return new_filename;
}

void FileVolume::
expand_bname() {
  // A FileVolume tries to create its bucket directory as soon as it
  // can.
  VolumeBase::expand_bname();
  if (exists()) {
    mkdir_p(_dirname);
  }
}

string FileVolume::
move_file(const string &old_filename, const string &basename, 
	  bool leave_orig) {
  string new_filename = _dirname + "/" + basename;
  string command = "mv -f ";
  if (leave_orig) {
    command = "cp -f ";
  }

  command += protect_from_shell(old_filename) + " " + 
    protect_from_shell(new_filename);

  int status = system(command.c_str());
  
  if (status < 0) {
    perror("system");
    return string();
  }
  
  if (status != 0) {
    cerr << "Unable to move into install directory.\n";
    return string();
  }

  if (_bname.empty()) {
    return basename;
  } else {
    return _bname + "/" + basename;
  }
}

bool FileVolume::
is_bucket_full() {
  int current_size;
  int num_files;
  stat_dir(_dirname, current_size, num_files);

  // If the bucket_size is 0, it means the bucket will never be
  // filled.
  
  // If the directory only contains one file, we will never consider
  // the bucket filled.  This prevents us from running away with the
  // system if the user specifes a too-small bucket size (smaller than
  // a single file).

  // Otherwise, the bucket is filled if we have reached or exceeded
  // the specified bucket size.

  return (_bucket_size > 0 &&
	  num_files > 1 &&
	  current_size >= _bucket_size);
}

void FileVolume::
increment_bucket() {
  music_db->sql_command
    ("UPDATE volumes"
     " SET current_bucket=" + DataCell(_bucket + 1) +
     " WHERE volume_id=" + DataCell(_volume_id));

  read_db();
}


// stat_dir() reads the given directory and sums up the total size of all
// files in it.
void FileVolume::  
stat_dir(const string &dirname, int &size_megabytes, int &num_files) {
  static const unsigned long megabyte = 1048576;

  size_megabytes = 0;
  num_files = 0;

  DIR *dirp = opendir(dirname.c_str());
  if (dirp==NULL) {
    perror(dirname.c_str());
    return;
  }

  string prefix = dirname + '/';
  unsigned long size_bytes = 0;

  struct stat statbuf;
  struct dirent *ent = readdir(dirp);
  while (ent!=NULL) {
    string filename = prefix + ent->d_name;
    if (stat(filename.c_str(), &statbuf) >= 0) {
      if (S_ISREG(statbuf.st_mode)) {
	size_bytes += statbuf.st_size;
	num_files++;
      }
    } else {
      perror(filename.c_str());
    }
    ent = readdir(dirp);
  }

  size_megabytes = (int)(size_bytes / megabyte);
}
